//
//	File:	addequal.C
//
//	Desc:	This application is intended to demonstrate the proper
//			steps in order to add a security equivalence to an object.
//			Please note that you have to have manage rights to both
//			objects in order to perform this operation.
//			Besides that, this sample contains some handy miscellaneous
//			functions to use when working with DS attributes.
//
//
//		DISCLAIMER
//
//	Novell, Inc. makes no representations or warranties with respect to
//	any NetWare software, and specifically disclaims any express or
//	implied warranties of merchantability, title, or fitness for a
//	particular purpose.  
//
//	Distribution of any NetWare software is forbidden without the
//	express written consent of Novell, Inc.  Further, Novell reserves
//	the right to discontinue distribution of any NetWare software.
//	
//	Novell is not responsible for lost profits or revenue, loss of use
//	of the software, loss of data, costs of re-creating lost data, the
//	cost of any substitute equipment or program, or claims by any party
//	other than you.  Novell strongly recommends a backup be made before
//	any software is installed.   Technical support for this software
//	may be provided at the discretion of Novell.
//
//	Programmers:
//
//		Ini	Who						Firm
//		-----------------------------------------------------------------------
//		HW		Hans Wieser				Novell Developer Support.
//
//	History:
//
//		When		Who	What
//		-----------------------------------------------------------------------
//		05-20-96	hw		First code.
//

//
//	Include headers, macros, function prototypes, etc.
//
// NetWare includes
#if defined(__WIN32__)
	#define	N_ARCH_32
	#define	N_PLAT_WNT
#elif defined(__OS2__)
	#define	N_PLAT_OS2
   #define	N_ARCH_32
#else
	#define	NWDOS
#endif
#include	<nwnet.h>
#include <nwcalls.h>
#define	SUCCESS(x)	((x) == N_SUCCESS)

// CRTL includes
#include <stdio.h>
#include <string.h>

// Global variables
static nstr8					equivalentToMe[] = "Equivalent To Me";
static nstr8					securityEquals[] = A_SECURITY_EQUALS;

// Helper functions
N_INTERN_FUNC(nbool)	CheckPrivileges(NWDSContextHandle, pnstr8, pnstr8, pnstr8, nuint32);
N_INTERN_FUNC(nbool)	AttributeExists(NWDSContextHandle, pnstr8, pnstr8);
N_INTERN_FUNC(nbool)	AddAttributeDNValue(NWDSContextHandle, pnstr8, pnstr8, pnstr8);

void
main(int argc, char** argv)	{
	static NWDSContextHandle	hContext;
	static NWCCODE					ccode;

	if(argc != 3)	{
		puts("Usage:\tAddEqual <DN1> <DN2>");
		puts("where\tDN1 is the object name to receive");
		puts("\tsecurity equivalence to the object named by DN2");
		puts("Example:\n\tAddEqual CN=Bob.O=Novell CN=Joe.OU=Sales.O=Novell");
		return;
	}

	if(!SUCCESS(ccode = NWCallsInit(NULL, NULL)))	{
		printf("Cannot initialize libraries. Error 0x%04X.\n", ccode);
		return;
	}
	if(!SUCCESS(ccode = NWDSCreateContextHandle(&hContext)))	{
		printf("Cannot create context. Error 0x%04X.\n", ccode);
		return;
	}
	// make sure we clean up properly
	__try	{
		nflag32	flags;
		nstr8		object[MAX_DN_CHARS];   // being granted rights
		nstr8		subject[MAX_DN_CHARS];  // being referenced
		nstr8		user[MAX_DN_CHARS];     // you

		// setup context to use typeful DN only
		if(!SUCCESS(ccode = NWDSGetContext(hContext, DCK_FLAGS, &flags)))	{
			printf("Error %d retrieving flags", ccode);
			return;
		}
		flags &= ~DCV_CANONICALIZE_NAMES;// Distinguished names
		flags &= ~DCV_TYPELESS_NAMES;    // Require typeful names explicitly
		if(!SUCCESS(ccode = NWDSSetContext(hContext, DCK_FLAGS, &flags)))	{
			printf("Error %d setting flags", ccode);
			return;
		}

		// get the user name
		if(!SUCCESS(ccode = NWDSWhoAmI(hContext, user)))	{
			printf("Fatal: Cannot get user name. Error %d.\n", ccode);
			return;
		}

		// make sure the names are valid
		if(!SUCCESS(ccode = NWDSCanonicalizeName(
			hContext,
			argv[1],
			object))
			||
			!SUCCESS(ccode = NWDSCanonicalizeName(
			hContext,
			argv[2],
			subject)))	{
			printf("Error %d validating input parameters.\n", ccode);
			return;
		}

		// Check necessary rights to perform operation
		if(!CheckPrivileges(
			hContext,
			user,
			subject,
			equivalentToMe,
			DS_ATTR_WRITE))	{
			printf("Insufficient rights\n\tfor %s \n\ton attribute %s\n\tof %s.\n",
				user,
				equivalentToMe,
				subject);
			return;
		}

		if(!CheckPrivileges(
			hContext,
			user,
			object,
			securityEquals,
			DS_ATTR_WRITE))	{
			printf("Insufficient rights\n\tfor %s \n\ton attribute %s\n\tof %s.\n",
				user,
				securityEquals,
				object);
			return;
		}

		if(!AddAttributeDNValue(
			hContext,
			subject,
			equivalentToMe,
			object)
			||
			!AddAttributeDNValue(
			hContext,
			object,
			securityEquals,
			subject))
			puts("Error adding attributes.");

		// we are authorized to perform this operation
	}
	__finally	{
		NWDSFreeContext(hContext);
	}
}

N_INTERN_FUNC(nbool)
CheckPrivileges(	NWDSContextHandle	hContext,
						pnstr8				objectName,
						pnstr8				subjectName,
						pnstr8				attributeName,
						nuint32				rightsMask)	{
	nuint32  rightsGranted;

	return SUCCESS(NWDSGetEffectiveRights(
		hContext,
		objectName,
		subjectName,
		attributeName,
		&rightsGranted))
		&&
		(rightsMask & rightsGranted);
}

N_INTERN_FUNC(nbool)
AttributeExists(	NWDSContextHandle	hContext,
						pnstr8				subjectName,
						pnstr8				attributeName)	{
	nbool	retVal = N_FALSE;

	pBuf_T	attributeFilter = NULL;
	pBuf_T	resultBuffer = NULL;
	nint32	iteration = NO_MORE_ITERATIONS;

	__try	{
		if(!SUCCESS(NWDSAllocBuf(DEFAULT_MESSAGE_LEN, &attributeFilter))
			||
			!SUCCESS(NWDSAllocBuf(DEFAULT_MESSAGE_LEN, &resultBuffer)))	{
			puts("Buffer allocation failure.");
			return retVal;
		}

		if(!SUCCESS(NWDSInitBuf(hContext, DSV_READ, attributeFilter))
			||
			!SUCCESS(NWDSPutAttrName(hContext, attributeFilter, attributeName)))	{
			puts("Buffer initialization failed.");
			return retVal;
		}
		retVal = SUCCESS(NWDSRead(
			hContext,
			subjectName,
			DS_EFFECTIVE_PRIVILEGES,
			N_FALSE,
			attributeFilter,
			&iteration,
			resultBuffer));
	}
	__finally	{
		if(attributeFilter != NULL)
			NWDSFreeBuf(attributeFilter);
		if(resultBuffer != NULL)
			NWDSFreeBuf(resultBuffer);
	}
	return retVal;
}

N_INTERN_FUNC(nbool)
AddAttributeDNValue(	NWDSContextHandle	hContext,
							pnstr8				subjectName,
							pnstr8				attributeName,
							pnstr8				attributeValue)	{

	pBuf_T	changes = NULL;
	nbool		retVal = N_FALSE;
	nint32	iteration = 0;
	NWCCODE	ccode;

	__try	{
		if(!SUCCESS(NWDSAllocBuf(DEFAULT_MESSAGE_LEN, &changes)))	{
			puts("Buffer allocation failed.");
			return retVal;
		}
		// initialize the input buffer
		if(!SUCCESS(ccode = NWDSInitBuf(hContext, DSV_MODIFY_ENTRY, changes)))	{
			printf("Could not initialize modification buffer. Error %d.\n", ccode);
			return retVal;
		}

		if(!AttributeExists(
			hContext,
			subjectName,
			attributeName))	{

			// attempt to add the attribute itself
			// after all it's just one more modification
			if(!SUCCESS(ccode = NWDSPutChange(
				hContext,
				changes,
				DS_ADD_ATTRIBUTE,
				attributeName)))	{
				printf("Error %d preparing addition of attribute %s to %s .\n",
					ccode,
					attributeName,
					subjectName);
				return retVal;
			}
		}

		// attribute should be there now
		if(!SUCCESS(ccode = NWDSPutChange(
			hContext,
			changes,
			DS_ADD_VALUE,
			attributeName))
			||
			!SUCCESS(ccode = NWDSPutAttrVal(
			hContext,
			changes,
			SYN_DIST_NAME,
			attributeValue)))	{
			printf("Error %d preparing value addition to %s of %s .\n",
				ccode,
				attributeName,
				subjectName);
			return retVal;
		}
		if(!SUCCESS(ccode = NWDSModifyObject(
			hContext,
			subjectName,
			&iteration,
			0,
			changes)))
			printf("Could not add value %s to attribute %s of %s . Error %d\n",
				attributeValue,
				attributeName,
				subjectName,
				ccode);
		else
			retVal = N_TRUE;
	}
	__finally	{
		if(changes != NULL)
			NWDSFreeBuf(changes);
	}
	return retVal;
}